@@ -7,52 +7,68 @@ import os
 import random
 import re
 import sys
+import glob
 import time
 from datetime import datetime
 from enum import Enum, auto
 from subprocess import call
+import typing
 
 import attr
 import click
 import pendulum
 import psutil
 
-from plotman import chia
+from plotman import chia, madmax
 
 
-def job_phases_for_tmpdir(d, all_jobs):
+def job_phases_for_tmpdir(d: str, all_jobs: typing.List["Job"]) -> typing.List["Phase"]:
     '''Return phase 2-tuples for jobs running on tmpdir d'''
-    return sorted([j.progress() for j in all_jobs if j.tmpdir == d])
+    return sorted([j.progress() for j in all_jobs if os.path.normpath(j.tmpdir) == os.path.normpath(d)])
 
-def job_phases_for_dstdir(d, all_jobs):
+def job_phases_for_dstdir(d: str, all_jobs: typing.List["Job"]) -> typing.List["Phase"]:
     '''Return phase 2-tuples for jobs outputting to dstdir d'''
-    return sorted([j.progress() for j in all_jobs if j.dstdir == d])
+    return sorted([j.progress() for j in all_jobs if os.path.normpath(j.dstdir) == os.path.normpath(d)])
 
-def is_plotting_cmdline(cmdline):
-    if cmdline and 'python' in cmdline[0].lower():
+def is_plotting_cmdline(cmdline: typing.List[str]) -> bool:
+    if cmdline and 'python' in cmdline[0].lower():  # Stock Chia plotter
         cmdline = cmdline[1:]
-    return (
-        len(cmdline) >= 3
-        and 'chia' in cmdline[0]
-        and 'plots' == cmdline[1]
-        and 'create' == cmdline[2]
-    )
+        return (
+            len(cmdline) >= 3
+            and 'chia' in cmdline[0]
+            and 'plots' == cmdline[1]
+            and 'create' == cmdline[2]
+        )
+    elif cmdline and 'chia_plot' == os.path.basename(cmdline[0].lower()):  # Madmax plotter
+        return True
+    return False
 
-def parse_chia_plot_time(s):
+def parse_chia_plot_time(s: str) -> pendulum.DateTime:
     # This will grow to try ISO8601 as well for when Chia logs that way
-    return pendulum.from_format(s, 'ddd MMM DD HH:mm:ss YYYY', locale='en', tz=None)
+    # TODO: unignore once fixed upstream
+    #       https://github.com/sdispater/pendulum/pull/548
+    return pendulum.from_format(s, 'ddd MMM DD HH:mm:ss YYYY', locale='en', tz=None)  # type: ignore[arg-type]
 
-def parse_chia_plots_create_command_line(command_line):
+def parse_chia_plots_create_command_line(
+    command_line: typing.List[str],
+) -> "ParsedChiaPlotsCreateCommand":
     command_line = list(command_line)
     # Parse command line args
-    if 'python' in command_line[0].lower():
+    if 'python' in command_line[0].lower():  # Stock Chia plotter
         command_line = command_line[1:]
-    assert len(command_line) >= 3
-    assert 'chia' in command_line[0]
-    assert 'plots' == command_line[1]
-    assert 'create' == command_line[2]
-
-    all_command_arguments = command_line[3:]
+        assert len(command_line) >= 3
+        assert 'chia' in command_line[0]
+        assert 'plots' == command_line[1]
+        assert 'create' == command_line[2]
+        all_command_arguments = command_line[3:]
+        # TODO: We could at some point do chia version detection and pick the
+        #       associated command.  For now we'll just use the latest one we have
+        #       copied.
+        command = chia.commands.latest_command()
+    elif 'chia_plot' in command_line[0].lower():  # Madmax plotter
+        command_line = command_line[1:]
+        all_command_arguments = command_line[2:]
+        command = madmax._cli_c8121b9
 
     # nice idea, but this doesn't include -h
     # help_option_names = command.get_help_option_names(ctx=context)
@@ -64,10 +80,6 @@ def parse_chia_plots_create_command_line(command_line):
         if argument not in help_option_names
     ]
 
-    # TODO: We could at some point do chia version detection and pick the
-    #       associated command.  For now we'll just use the latest one we have
-    #       copied.
-    command = chia.commands.latest_command()
     try:
         context = command.make_context(info_name='', args=list(command_arguments))
     except click.ClickException as e:
@@ -84,7 +96,12 @@ def parse_chia_plots_create_command_line(command_line):
     )
 
 class ParsedChiaPlotsCreateCommand:
-    def __init__(self, error, help, parameters):
+    def __init__(
+        self,
+        error: click.ClickException,
+        help: bool,
+        parameters: typing.Dict[str, object],
+    ) -> None:
         self.error = error
         self.help = help
         self.parameters = parameters
@@ -96,14 +113,14 @@ class Phase:
     minor: int = 0
     known: bool = True
 
-    def __lt__(self, other):
+    def __lt__(self, other: "Phase") -> bool:
         return (
             (not self.known, self.major, self.minor)
             < (not other.known, other.major, other.minor)
         )
 
     @classmethod
-    def from_tuple(cls, t):
+    def from_tuple(cls, t: typing.Tuple[typing.Optional[int], typing.Optional[int]]) -> "Phase":
         if len(t) != 2:
             raise Exception(f'phase must be created from 2-tuple: {t!r}')
 
@@ -113,28 +130,50 @@ class Phase:
         if t[0] is None:
             return cls(known=False)
 
-        return cls(major=t[0], minor=t[1])
+        return cls(major=t[0], minor=t[1])  # type: ignore[arg-type]
 
     @classmethod
-    def list_from_tuples(cls, l):
+    def list_from_tuples(
+        cls,
+        l: typing.Sequence[typing.Tuple[typing.Optional[int], typing.Optional[int]]],
+    ) -> typing.List["Phase"]:
         return [cls.from_tuple(t) for t in l]
 
+    def __str__(self) -> str:
+        if not self.known:
+            return '?:?'
+        return f'{self.major}:{self.minor}'
+
 # TODO: be more principled and explicit about what we cache vs. what we look up
 # dynamically from the logfile
 class Job:
     'Represents a plotter job'
 
-    logfile = ''
-    jobfile = ''
-    job_id = 0
-    plot_id = '--------'
-    proc = None   # will get a psutil.Process
+    logfile: str = ''
+    jobfile: str = ''
+    job_id: int = 0
+    plot_id: str = '--------'
+    plotter: str = ''
+    proc: psutil.Process
+    k: int
+    r: int
+    u: int
+    b: int
+    n: int
+    tmpdir: str
+    tmp2dir: str
+    dstdir: str
 
-    def get_running_jobs(logroot, cached_jobs=()):
+    @classmethod
+    def get_running_jobs(
+        cls,
+        logroot: str,
+        cached_jobs: typing.Sequence["Job"] = (),
+    ) -> typing.List["Job"]:
         '''Return a list of running plot jobs.  If a cache of preexisting jobs is provided,
            reuse those previous jobs without updating their information.  Always look for
            new jobs not already in the cache.'''
-        jobs = []
+        jobs: typing.List[Job] = []
         cached_jobs_by_pid = { j.proc.pid: j for j in cached_jobs }
 
         with contextlib.ExitStack() as exit_stack:
@@ -183,7 +222,7 @@ class Job:
                             )
                             if parsed_command.error is not None:
                                 continue
-                            job = Job(
+                            job = cls(
                                 proc=proc,
                                 parsed_command=parsed_command,
                                 logroot=logroot,
@@ -195,7 +234,12 @@ class Job:
         return jobs
 
 
-    def __init__(self, proc, parsed_command, logroot):
+    def __init__(
+        self,
+        proc: psutil.Process,
+        parsed_command: ParsedChiaPlotsCreateCommand,
+        logroot: str,
+    ) -> None:
         '''Initialize from an existing psutil.Process object.  must know logroot in order to understand open files'''
         self.proc = proc
         # These are dynamic, cached, and need to be udpated periodically
@@ -224,17 +268,26 @@ class Job:
         #     'nobitfield': False,
         #     'exclude_final_dir': False,
         # }
+        if proc.name().startswith("chia_plot"): # MADMAX
+            self.k = 32
+            self.r = self.args['threads']  # type: ignore[assignment]
+            self.u = self.args['buckets']  # type: ignore[assignment]
+            self.b = 0
+            self.n = self.args['count']  # type: ignore[assignment]
+            self.tmpdir = self.args['tmpdir']  # type: ignore[assignment]
+            self.tmp2dir = self.args['tmpdir2']  # type: ignore[assignment]
+            self.dstdir = self.args['finaldir']  # type: ignore[assignment]
+        else: # CHIA
+            self.k = self.args['size']  # type: ignore[assignment]
+            self.r = self.args['num_threads']  # type: ignore[assignment]
+            self.u = self.args['buckets']  # type: ignore[assignment]
+            self.b = self.args['buffer']  # type: ignore[assignment]
+            self.n = self.args['num']  # type: ignore[assignment]
+            self.tmpdir = self.args['tmp_dir']  # type: ignore[assignment]
+            self.tmp2dir = self.args['tmp2_dir']  # type: ignore[assignment]
+            self.dstdir = self.args['final_dir']  # type: ignore[assignment]
 
-        self.k = self.args['size']
-        self.r = self.args['num_threads']
-        self.u = self.args['buckets']
-        self.b = self.args['buffer']
-        self.n = self.args['num']
-        self.tmpdir = self.args['tmp_dir']
-        self.tmp2dir = self.args['tmp2_dir']
-        self.dstdir = self.args['final_dir']
-
-        plot_cwd = self.proc.cwd()
+        plot_cwd: str = self.proc.cwd()
         self.tmpdir = os.path.join(plot_cwd, self.tmpdir)
         if self.tmp2dir is not None:
             self.tmp2dir = os.path.join(plot_cwd, self.tmp2dir)
@@ -262,7 +315,7 @@ class Job:
 
 
 
-    def init_from_logfile(self):
+    def init_from_logfile(self) -> None:
         '''Read plot ID and job start time from logfile.  Return true if we
            find all the info as expected, false otherwise'''
         assert self.logfile
@@ -275,11 +328,22 @@ class Job:
                 with contextlib.suppress(UnicodeDecodeError):
                     for line in f:
                         m = re.match('^ID: ([0-9a-f]*)', line)
-                        if m:
+                        if m: # CHIA
                             self.plot_id = m.group(1)
+                            self.plotter = 'chia'
                             found_id = True
+                        else: 
+                            m = re.match(r"^Plot Name: plot-k(\d+)-(\d+)-(\d+)-(\d+)-(\d+)-(\d+)-(\w+)$", line)
+                            if m: # MADMAX
+                                self.plot_id = m.group(7)
+                                self.plotter = 'madmax'
+                                self.start_time = pendulum.from_timestamp(os.path.getctime(self.logfile))
+                                found_id = True
+                                found_log = True
+                                break
+
                         m = re.match(r'^Starting phase 1/4:.*\.\.\. (.*)', line)
-                        if m:
+                        if m: # CHIA
                             # Mon Nov  2 08:39:53 2020
                             self.start_time = parse_chia_plot_time(m.group(1))
                             found_log = True
@@ -295,15 +359,15 @@ class Job:
         # TODO: we never come back to this; e.g. plot_id may remain uninitialized.
         # TODO: should we just use the process start time instead?
         if not found_log:
-            self.start_time = datetime.fromtimestamp(os.path.getctime(self.logfile))
+            self.start_time = pendulum.from_timestamp(os.path.getctime(self.logfile))
 
         # Load things from logfile that are dynamic
         self.update_from_logfile()
 
-    def update_from_logfile(self):
+    def update_from_logfile(self) -> None:
         self.set_phase_from_logfile()
 
-    def set_phase_from_logfile(self):
+    def set_phase_from_logfile(self) -> None:
         assert self.logfile
 
         # Map from phase number to subphase number reached in that phase.
@@ -316,27 +380,63 @@ class Job:
         with open(self.logfile, 'r') as f:
             with contextlib.suppress(UnicodeDecodeError):
                 for line in f:
-                    # "Starting phase 1/4: Forward Propagation into tmp files... Sat Oct 31 11:27:04 2020"
-                    m = re.match(r'^Starting phase (\d).*', line)
-                    if m:
-                        phase = int(m.group(1))
-                        phase_subphases[phase] = 0
+                    if self.plotter == "madmax":
 
-                    # Phase 1: "Computing table 2"
-                    m = re.match(r'^Computing table (\d).*', line)
-                    if m:
-                        phase_subphases[1] = max(phase_subphases[1], int(m.group(1)))
+                        # MADMAX reports after completion of phases so increment the reported subphases
+                        # and assume that phase 1 has already started
 
-                    # Phase 2: "Backpropagating on table 2"
-                    m = re.match(r'^Backpropagating on table (\d).*', line)
-                    if m:
-                        phase_subphases[2] = max(phase_subphases[2], 7 - int(m.group(1)))
+                        # MADMAX: "[P1]" or "[P2]" or "[P4]"
+                        m = re.match(r'^\[P(\d)\].*', line)
+                        if m:
+                            phase = int(m.group(1))
+                            phase_subphases[phase] = 1
 
-                    # Phase 3: "Compressing tables 4 and 5"
-                    m = re.match(r'^Compressing tables (\d) and (\d).*', line)
-                    if m:
-                        phase_subphases[3] = max(phase_subphases[3], int(m.group(1)))
+                        # MADMAX: "[P1] or [P2] Table 7"
+                        m = re.match(r'^\[P(\d)\] Table (\d).*', line)
+                        if m:
+                            phase = int(m.group(1))
+                            if phase == 1:
+                                phase_subphases[1] = max(phase_subphases[1], (int(m.group(2))+1))
 
+                            elif phase == 2:
+                                if 'rewrite' in line:
+                                    phase_subphases[2] = max(phase_subphases[2], (9 - int(m.group(2))))
+                                else:
+                                    phase_subphases[2] = max(phase_subphases[2], (8 - int(m.group(2))))
+
+                        # MADMAX: Phase 3: "[P3-1] Table 4"
+                        m = re.match(r'^\[P3\-(\d)\] Table (\d).*', line)
+                        if m:
+                            if 3 in phase_subphases:
+                                if int(m.group(1)) == 2:
+                                    phase_subphases[3] = max(phase_subphases[3], int(m.group(2)))
+                                else:
+                                    phase_subphases[3] = max(phase_subphases[3], int(m.group(2))-1)
+                            else: 
+                                phase_subphases[3] = 1
+
+                    else:                    
+                        # CHIA: "Starting phase 1/4: Forward Propagation into tmp files... Sat Oct 31 11:27:04 2020"
+                        m = re.match(r'^Starting phase (\d).*', line)
+                        if m:
+                            phase = int(m.group(1))
+                            phase_subphases[phase] = 0
+                        
+                        # CHIA: Phase 1: "Computing table 2"
+                        m = re.match(r'^Computing table (\d).*', line)
+                        if m:
+                            phase_subphases[1] = max(phase_subphases[1], int(m.group(1)))
+                        
+                        # CHIA: Phase 2: "Backpropagating on table 2"
+                        m = re.match(r'^Backpropagating on table (\d).*', line)
+                        if m:
+                            phase_subphases[2] = max(phase_subphases[2], 7 - int(m.group(1)))
+
+                        # CHIA: Phase 3: "Compressing tables 4 and 5"
+                        m = re.match(r'^Compressing tables (\d) and (\d).*', line)
+                        if m:
+                            phase_subphases[3] = max(phase_subphases[3], int(m.group(1)))
+                    
                     # TODO also collect timing info:
 
                     # "Time for phase 1 = 22796.7 seconds. CPU (98%) Tue Sep 29 17:57:19 2020"
@@ -355,15 +455,15 @@ class Job:
         else:
             self.phase = Phase(major=0, minor=0)
 
-    def progress(self):
+    def progress(self) -> Phase:
         '''Return a 2-tuple with the job phase and subphase (by reading the logfile)'''
         return self.phase
 
-    def plot_id_prefix(self):
+    def plot_id_prefix(self) -> str:
         return self.plot_id[:8]
 
     # TODO: make this more useful and complete, and/or make it configurable
-    def status_str_long(self):
+    def status_str_long(self) -> str:
         return '{plot_id}\nk={k} r={r} b={b} u={u}\npid:{pid}\ntmp:{tmp}\ntmp2:{tmp2}\ndst:{dst}\nlogfile:{logfile}'.format(
             plot_id = self.plot_id,
             k = self.k,
@@ -374,14 +474,49 @@ class Job:
             tmp = self.tmpdir,
             tmp2 = self.tmp2dir,
             dst = self.dstdir,
-            plotid = self.plot_id,
             logfile = self.logfile
             )
 
-    def get_mem_usage(self):
-        return self.proc.memory_info().vms  # Total, inc swapped
+    def print_logs(self, follow: bool = False) -> None:
+        with open(self.logfile, 'r') as f:
+            if follow:
+                line = ''
+                while True:
+                    tmp = f.readline()
+                    if tmp is not None:
+                        line += tmp
+                        if line.endswith("\n"):
+                            print(line.rstrip('\n'))
+                            line = ''
+                    else:
+                        time.sleep(0.1)
+            else:
+                print(f.read())
 
-    def get_tmp_usage(self):
+    def to_dict(self) -> typing.Dict[str, object]:
+        '''Exports important information as dictionary.'''
+        return dict(
+            plot_id=self.plot_id[:8],
+            k=self.k,
+            tmp_dir=self.tmpdir,
+            dst_dir=self.dstdir,
+            progress=str(self.progress()),
+            tmp_usage=self.get_tmp_usage(),
+            pid=self.proc.pid,
+            run_status=self.get_run_status(),
+            mem_usage=self.get_mem_usage(),
+            time_wall=self.get_time_wall(),
+            time_user=self.get_time_user(),
+            time_sys=self.get_time_sys(),
+            time_iowait=self.get_time_iowait()
+        )
+
+
+    def get_mem_usage(self) -> int:
+        # Total, inc swapped
+        return self.proc.memory_info().vms  # type: ignore[no-any-return]
+
+    def get_tmp_usage(self) -> int:
         total_bytes = 0
         with contextlib.suppress(FileNotFoundError):
             # The directory might not exist at this name, or at all, anymore
@@ -393,7 +528,7 @@ class Job:
                             total_bytes += entry.stat().st_size
         return total_bytes
 
-    def get_run_status(self):
+    def get_run_status(self) -> str:
         '''Running, suspended, etc.'''
         status = self.proc.status()
         if status == psutil.STATUS_RUNNING:
@@ -405,19 +540,19 @@ class Job:
         elif status == psutil.STATUS_STOPPED:
             return 'STP'
         else:
-            return self.proc.status()
+            return self.proc.status()  # type: ignore[no-any-return]
 
-    def get_time_wall(self):
+    def get_time_wall(self) -> int:
         create_time = datetime.fromtimestamp(self.proc.create_time())
         return int((datetime.now() - create_time).total_seconds())
 
-    def get_time_user(self):
+    def get_time_user(self) -> int:
         return int(self.proc.cpu_times().user)
 
-    def get_time_sys(self):
+    def get_time_sys(self) -> int:
         return int(self.proc.cpu_times().system)
 
-    def get_time_iowait(self):
+    def get_time_iowait(self) -> typing.Optional[int]:
         cpu_times = self.proc.cpu_times()
         iowait = getattr(cpu_times, 'iowait', None)
         if iowait is None:
@@ -425,26 +560,24 @@ class Job:
 
         return int(iowait)
 
-    def suspend(self, reason=''):
+    def suspend(self, reason: str = '') -> None:
         self.proc.suspend()
         self.status_note = reason
 
-    def resume(self):
+    def resume(self) -> None:
         self.proc.resume()
 
-    def get_temp_files(self):
+    def get_temp_files(self) -> typing.Set[str]:
         # Prevent duplicate file paths by using set.
         temp_files = set([])
-        for f in self.proc.open_files():
-            if any(
-                dir in f.path
-                for dir in [self.tmpdir, self.tmp2dir, self.dstdir]
-                if dir is not None
-            ):
-                temp_files.add(f.path)
+
+        for dir in [self.tmpdir, self.tmp2dir, self.dstdir]:
+            if dir is not None:
+                temp_files.update(glob.glob(os.path.join(dir, f"plot-*-{self.plot_id}.tmp")))
+
         return temp_files
 
-    def cancel(self):
+    def cancel(self) -> None:
         'Cancel an already running job'
         # We typically suspend the job as the first action in killing it, so it
         # doesn't create more tmp files during death.  However, terminate() won't
